module gradient4d_mod

  use sep
  use omp_lib

  implicit none

  integer, private                           :: n1, n2, nt, ntau, otau, it, itau, tau, i1, i2
  real, private                              :: idt
  real, private                              :: b1x, b2x, b3x, b4x, b1z, b2z, b3z, b4z
  real, allocatable, dimension(:), private   :: uaux

  contains

!!--------------------------------------------------------------------------
  subroutine grad_init(n1_in, d1, n2_in, d2)
    integer         :: n1_in, n2_in
    real            :: d1, d2
    n1 = n1_in
    n2 = n2_in
    b1x = ( +1224./1024.)/d2
    b2x = ( -245./3072.)/d2
    b3x = ( +49./5120.)/d2
    b4x = ( -5./7168.)/d2
    b1z = ( +1224./1024.)/d1
    b2z = ( -245./3072.)/d1
    b3z = ( +49./5120.)/d1
    b4z = ( -5./7168.)/d1
    allocate(uaux(max(n1,n2)))
  end subroutine


  subroutine grad(data, model)
    !! 8th order accuracy first derivative
    real, dimension(:,:)    :: data
    real, dimension(:,:,:)  :: model
  
    model=0
  !  if (adj) then    !! RTM
!$omp parallel do default(shared) &
!$omp private(i1,i2)
      do i2=4,n2-4
        do i1=4,n1-4
          model(2,i1,i2) =  b1x*(data(i1  ,i2+1) - data(i1  ,i2  )) &
                          + b2x*(data(i1  ,i2+2) - data(i1  ,i2-1)) &
                          + b3x*(data(i1  ,i2+3) - data(i1  ,i2-2)) &
                          + b4x*(data(i1  ,i2+4) - data(i1  ,i2-3))
  
          model(1,i1,i2) =  b1z*(data(i1+1,i2  ) - data(i1  ,i2  )) &
                          + b2z*(data(i1+2,i2  ) - data(i1-1,i2  )) & 
                          + b3z*(data(i1+3,i2  ) - data(i1-2,i2  )) &
                          + b4z*(data(i1+4,i2  ) - data(i1-3,i2  )) 
        end do
      end do
!$omp end parallel do
  !  else 
  !  end if
  
!omp parallel do default(shared) &
!omp private(i2, uaux)
    do i2 = 1,n2
      CALL interp1d(model(1,:,i2),uaux(1:n1),.FALSE.)
      model(1,:,i2)=uaux(1:n1)
    end do 
!omp end parallel do

!omp parallel do default(shared) &
!omp private(i1, uaux)
    do i1 = 1,n1
      CALL interp1d(model(2,i1,:),uaux(1:n2),.FALSE.)
      model(2,i1,:)=uaux(1:n2)
    end do 
!omp end parallel do
  end subroutine
  
  SUBROUTINE interp1d(model,data,adj)
  !*** I borrowed this subroutine to interpolate from staggered grid to regular grid
    INTEGER :: jm
    INTEGER :: jd
  
    REAL :: fx
    REAL :: gx
  
    REAL, DIMENSION(:) :: data
    REAL, DIMENSION(:) :: model
  
    LOGICAL :: adj
  
    IF(adj) THEN
      model=0
    ELSE
      data=0
    END IF
    gx=0.5
    fx=0.5
    DO jd=1,SIZE(data)
      jm=jd-1
      IF(jm > 1 .AND. jm <= SIZE(model)) THEN
        IF(adj) THEN
          model(jm)=  model(jm)  +gx*data(jd)
          model(jm+1)=model(jm+1)+fx*data(jd)
        ELSE
          data(jd)=data(jd)+gx*model(jm)+fx*model(jm+1)
        END IF
      END IF
    END DO
  END SUBROUTINE

  subroutine firstder_init(n1_in, n2_in, nt_in, dt_in)
    integer                      :: n1_in, n2_in, nt_in
    real                         :: dt_in
    n1 = n1_in
    n2 = n2_in
    nt = nt_in
    idt = 1./dt_in
  end subroutine

  function firstder_op(adj, add, model, data) result(stat)
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data
    integer            :: stat
    call firstder_op2(adj, add, model, data)
    stat=0
  end function

  subroutine firstder_op2(adj, add, model, data)
    logical,intent(in)          :: adj, add 
    real, dimension(n1,n2,nt) :: model, data
    if(adj) then
      data = data*idt
      if(.not. add) model = 0.
      do it = 1,nt-1
        model(:,:,it+1) =  model(:,:,it+1) + data(:,:,it)
        model(:,:,it)   =  model(:,:,it)   - data(:,:,it)
      end do
    else
      model = model*idt
      if(.not. add) data = 0.
!omp parallel do default(shared) &
!omp private(it)
      do it = 1,nt-1
        data(:,:,it) = data(:,:,it) -model(:,:,it)+model(:,:,it+1)
      end do
!omp end parallel do
    end if
  end subroutine

!!--------------------------------------------------------------------------

end module
